/*
 * ${license.title}
 * Copyright (C) 2022 ${license.owner}
 * ${license.mailto}
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include <filesystem>
#include <iostream>
#include <memory>

#include "fggl/app.hpp"
#include "fggl/audio/openal/audio.hpp"

#include "fggl/gfx/atlas.hpp"
#include "fggl/gfx/window.hpp"
#include "fggl/gfx/compat.hpp"
#include "fggl/gfx/ogl/compat.hpp"

#include "fggl/data/storage.hpp"
#include "fggl/util/service.h"

#include "fggl/ecs3/types.hpp"
#include "fggl/phys/bullet/bullet.hpp"

#include "fggl/scenes/menu.hpp"

#include "GameScene.h"
#include "rollball.hpp"

// prototype of resource discovery
void discover(const std::filesystem::path& base) {

	std::vector< std::filesystem::path > contentPacks;
	for ( const auto& item : std::filesystem::directory_iterator(base) ) {

		// content pack detection
		if ( std::filesystem::is_directory( item ) ) {
			auto manifest = item.path() / "manifest.yml";
			if ( std::filesystem::exists( manifest ) ) {
				contentPacks.push_back( item.path() );
			}
		}
		
	}

	// what did we find?
	std::cerr << "found pack(s): " << std::endl;
	for ( auto& pack : contentPacks ) {
		std::cerr << pack << std::endl;
	}

}

static void test_atlas_api() {
	// atlas testing
	std::vector< fggl::gfx::ImageAtlas<char>::SubImage > images;
	auto *atlas = fggl::gfx::ImageAtlas<char>::pack(images);
}

static void setup_service_locators(fggl::util::ServiceLocator& locator) {
	// FIXME: janky API(s)
	auto inputs = locator.supply<fggl::input::Input>(std::make_shared<fggl::input::Input>());
	auto storage = locator.supply<fggl::data::Storage>(std::make_shared<fggl::data::Storage>());

	locator.supply<fggl::gui::FontLibrary>(std::make_shared<fggl::gui::FontLibrary>());
	locator.supply<fggl::ecs3::TypeRegistry>(std::make_shared<fggl::ecs3::TypeRegistry>());
	locator.supply<fggl::audio::AudioService>(std::make_shared<fggl::audio::openAL>());
}

int main(int argc, const char* argv[]) {

	auto& locator = fggl::util::ServiceLocator::instance();
	setup_service_locators(locator);

	// create the application
	fggl::App app( "fggl-demo" );

	// Would be nice to not take args like this, it messes with lifetimes
	auto& windowing = app.use<fggl::gfx::ecsGlfwModule>(locator.get<fggl::input::Input>());

    // -- should not be our problem - this is a broken api
    auto window = windowing.createWindow("Demo Game");
    window->make_graphics<fggl::gfx::OpenGL4>();
    window->fullscreen( true );
    app.setWindow( std::move(window) );

	// load a bunch of modules to provide game functionality
	//app.use<fggl::ecs3::ecsTypes>();
	app.use<fggl::gfx::SceneUtils>();
	app.use<fggl::phys::Bullet3>();

	test_atlas_api();

	// Add a basic main menu
    auto *menu = app.add_state<fggl::scenes::BasicMenu>("menu");

	// add some menu items for the game states
	auto audio = locator.get<fggl::audio::AudioService>();
    menu->add("terrain", [&app, &audio]() {
		audio->play("click.ogg", false);
		app.change_state("game");
	});
	menu->add("rollball", [&app, &audio]() {
		audio->play("click.ogg", false);
		app.change_state("rollball");
	});
	menu->add("quit", [&app, &audio]() {
		audio->play("click.ogg", false);
		app.running(false);
	});

	// the game states themselves
    app.add_state<GameScene>("game");
	app.add_state<demo::RollBall>("rollball");

	return app.run(argc, argv);
}
